#include "particles.h"
#include <cmath>
#include <iostream>
#include <surface_mesh/surface_mesh.h>

using namespace std;
using namespace Eigen;
using namespace surface_mesh;

void makeGrid(ParticleSystem *psys);

void baseForce(ParticleSystem *psys)
{
    psys->forces.push_back(new DragForce(psys, 0.01));
    psys->forces.push_back(new GravityForce(psys, Vector3d(0, -9.807, 0)));
}

//***************************
//***** Particle System *****
//***************************

void ParticleSystem::init()
{
    makeGrid(this); // also creates spring forces
    baseForce(this);

    // OpenGL allocation
    glGenVertexArrays(2, _vao);
    glGenBuffers(2, _vbo);

    updateGL(true);
}

void ParticleSystem::init(const Mesh * mesh)
{
    LoadMesh(mesh);
    baseForce(this);

    // OpenGL allocation
    glGenVertexArrays(2, _vao);
    glGenBuffers(2, _vbo);

    updateGL(true);
}


void ParticleSystem::LoadMesh(const Mesh *mesh)
{

    const Surface_mesh he = mesh->halfEdgeMesh();
    const int nbVertex = he.n_vertices();
    const double mass = 1;                    // total mass
    const double pmass = mass / nbVertex;    // mass per particle
    const double k = 200.0, d = 0.1;        // spring constant, damping

    // create particles
    const Surface_mesh::Vertex_property<Point> vertices = he.get_vertex_property<Point>("v:point");
    const Vector3d v(0, 0, 0);

    for (Surface_mesh::Vertex_iterator it = he.vertices_begin(); it != he.vertices_end(); ++it) {
        Vector3d x(vertices[*it].x(), vertices[*it].y(), vertices[*it].z());
        particles.push_back(new Particle(pmass, x, v));
    }

    // create springs
    for (Surface_mesh::Edge_iterator it = he.edges_begin(); it != he.edges_end(); ++it) {
        const Surface_mesh::Vertex v0 = he.vertex(*it, 0), v1 = he.vertex(*it, 1);
        const Vector3f P0 = vertices[v0], P1 = vertices[v1];
        const double dist = (P1 - P0).norm();    //Longueur de l'arête
        forces.push_back(new SpringForce(particles[v0.idx()], particles[v1.idx()], k, d, dist));
    }
}


void ParticleSystem::UpdateMesh(Mesh &mesh)
{
    Surface_mesh he = mesh.halfEdgeMesh();
    const int nbVertex = he.n_vertices();
    assert(nbVertex == particles.size());

    auto pos = mesh.positions();
    for (int i = 0; i < particles.size(); ++i){
        Vector3d p = particles[i]->x;
        pos.col(i) = Vector3f(p.x(), p.y(), p.z());
    }
    mesh.updateAll();
    updateGL(false);
}

void ParticleSystem::updateGL(bool first)
{
    // Particles
    vector<Vector3f> positions;
    positions.resize(particles.size());
    for (int i = 0; i < particles.size(); i++) {
        positions[i] = particles[i]->x.cast<float>();
        _bbox.extend(positions[i]);
    }
    glBindVertexArray(_vao[0]);
    glBindBuffer(GL_ARRAY_BUFFER, _vbo[0]);
    if (first)
        glBufferData(GL_ARRAY_BUFFER, sizeof(Vector3f) * particles.size(), positions.data(), GL_DYNAMIC_DRAW);
    else
        glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(Vector3f) * particles.size(), positions.data());
    glBindVertexArray(0);

    // Spring forces
    positions.resize(0);
    positions.reserve(forces.size() * 2);
    for (int i = 0; i < forces.size(); i++) {
        SpringForce *s = dynamic_cast<SpringForce *>(forces[i]);
        if (s) {
            positions.push_back(s->p0->x.cast<float>());
            positions.push_back(s->p1->x.cast<float>());
        }
        AnchorForce *a = dynamic_cast<AnchorForce *>(forces[i]);
        if (a && a->p != nullptr) {
            positions.push_back(a->p->x.cast<float>());
            positions.push_back(a->x.cast<float>());
        }
    }

    glBindVertexArray(_vao[1]);
    glBindBuffer(GL_ARRAY_BUFFER, _vbo[1]);
    if (first || nbSpringForces != positions.size()) {
        nbSpringForces = positions.size();
        glBufferData(GL_ARRAY_BUFFER, sizeof(Vector3f) * nbSpringForces, positions.data(), GL_DYNAMIC_DRAW);
    } else
        glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(Vector3f) * nbSpringForces, positions.data());
    glBindVertexArray(0);
    _ready = true;
}

ParticleSystem::~ParticleSystem()
{
    glDeleteBuffers(2, _vbo);
    glDeleteVertexArrays(2, _vao);
}

void ParticleSystem::display(Shader *shader)
{
    if (!_ready)
        init();

    // Draw particles
    glBindVertexArray(_vao[0]);
    glBindBuffer(GL_ARRAY_BUFFER, _vbo[0]);

    GLuint vertex_loc = shader->getAttribLocation("vtx_position");
    glVertexAttribPointer(vertex_loc, 3, GL_FLOAT, GL_FALSE, 0, (void *) 0);
    glEnableVertexAttribArray(vertex_loc);

    glPointSize(5.f);
    glDrawArrays(GL_POINTS, 0, particles.size());

    glDisableVertexAttribArray(vertex_loc);
    glBindVertexArray(0);

    // Draw spring forces
    glBindVertexArray(_vao[1]);
    glBindBuffer(GL_ARRAY_BUFFER, _vbo[1]);

    vertex_loc = shader->getAttribLocation("vtx_position");
    glVertexAttribPointer(vertex_loc, 3, GL_FLOAT, GL_FALSE, 0, (void *) 0);
    glEnableVertexAttribArray(vertex_loc);

    glLineWidth(2.f);
    glDrawArrays(GL_LINES, 0, nbSpringForces * 2);

    glDisableVertexAttribArray(vertex_loc);
    glBindVertexArray(0);
}

void ParticleSystem::step(double dt)
{

    explicitEulerStep(this, dt);

    for(int p = 0; p < particles.size(); ++p) {
        for (int c = 0; c < colliders.size(); ++c) {
            CollisionResponse coll;

            if (colliders[c]->collide(particles[p], coll)) {
                particles[p]->v = coll.v;
                particles[p]->f += coll.f;
            }
        }
    }
}

int ParticleSystem::getDimensions()
{
    return static_cast<int>(6 * particles.size());
}

void ParticleSystem::getState(VectorXd &state)
{
    const int dim = getDimensions();
    state.resize(dim);
    for (int i = 0; i < particles.size(); ++i) {
        const int begin = i * 6;
        state.segment(begin, 3) = particles[i]->x;
        state.segment(begin + 3, 3) = particles[i]->v;
    }
}

void ParticleSystem::setState(const VectorXd &state)
{
    assert((state.size() / 6) != particles.size());
    for (int i = 0; i < particles.size(); ++i) {
        const int begin = i * 6;
        particles[i]->x = state.segment(begin, 3);
        particles[i]->v = state.segment(begin + 3, 3);
    }
}

void ParticleSystem::getDerivative(VectorXd &deriv)
{
    const int dim = getDimensions();
    deriv.resize(dim);
    for (Particle *p : particles) p->f.setZero();
    for (Force *force: forces) force->addForces();

    for (int i = 0; i < particles.size(); ++i) {
        const int begin = i * 6;
        deriv.segment(begin, 3) = particles[i]->v;
        deriv.segment(begin + 3, 3) = particles[i]->f / particles[i]->m;
    }
}

//*********************
//***** Add Force *****
//*********************
void GravityForce::addForces()
{
    if (ps->particles.empty()) return;
    for (Particle *p : ps->particles)
        p->f += p->m * g;
}

void DragForce::addForces()
{
    if (ps->particles.empty()) return;
    for (Particle *p : ps->particles)
        p->f += -1 * kd * p->v;
}

void SpringForce::addForces()
{
    // Force de 0 vers 1
    Vector3d l = p1->x - p0->x;
    const double l_norm = l.norm();
    const Vector3d f = (-1 * ks * (l_norm - l0) * l / l_norm) + (-1 * kd * ((p1->v - p0->v).dot(l)/l_norm) * l / l_norm);
    p1->f += f;
    p0->f -= f;
}

void AnchorForce::addForces()
{
    if (p == nullptr) return;
    Vector3d dx = p->x - x;
    p->f += -ks * dx - kd * p->v;
}

void makeGrid(ParticleSystem *psys)
{
    const int m = 10, n = 10;                                // resolution of system
    const double mass = 1;                                    // total mass
    const double pmass = mass / ((m + 1) * (n + 1));        // mass per particle
    const double x0 = 0.3, x1 = 0.7, y0 = 0.3, y1 = 0.7;    // extent of rectangle
    const double dx = (x1 - x0) / m, dy = (y1 - y0) / n;    // lengths of springs
    const double k = 200.0, d = 0.1;                        // spring constant, damping
    const double k_diag = k / 4, dy_diag = sqrt(dx * dx + dy * dy);    // spring constant, damping

    // create particles
    Particle *particles[m + 1][n + 1];
    for (int i = 0; i <= m; i++) {
        for (int j = 0; j <= n; j++) {
            Vector3d x(x0 + dx * i, y0 + dy * j, 0);
            Vector3d v(0, 0, 0);
            particles[i][j] = new Particle(pmass, x, v);
            psys->particles.push_back(particles[i][j]);
        }
    }

    // create anchor points
    //psys->forces.push_back(new AnchorForce(particles[0][n], particles[0][n]->x, 1000, 0.1));
    //psys->forces.push_back(new AnchorForce(particles[m][n], particles[m][n]->x, 1000, 0.1));

    // create springs
    for (int i = 0; i <= m; i++) {
        for (int j = 0; j <= n; j++) {
            Particle *p0 = particles[i][j];
            if (i < m) {
                Particle *p1 = particles[i + 1][j];
                psys->forces.push_back(new SpringForce(p0, p1, k, d, dx));
            }
            if (j < n) {
                Particle *p1 = particles[i][j + 1];
                psys->forces.push_back(new SpringForce(p0, p1, k, d, dy));
            }
            if (i < m && j < n) {
                Particle *p1 = particles[i + 1][j + 1];
                psys->forces.push_back(new SpringForce(p0, p1, k_diag, d, dy_diag));

            }
            if (i < m && j > 0) {
                Particle *p1 = particles[i + 1][j - 1];
                psys->forces.push_back(new SpringForce(p0, p1, k_diag, d, dy_diag));

            }
        }
    }
}
